import os
import sys
try:
    from file_archive.entry_point import entry_point
except ImportError:
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    from file_archive.entry_point import entry_point


if __name__ == '__main__':
    entry_point()



"""Python 2/3 compatibility symbols.

isinstance() checks:
 * Use string_types instead of 2's basestring
 * Use int_types instead of 3's int

sha1:
 * Silently accepts unicode so long as it's ASCII

BytesIO, StringIO
"""

from __future__ import division, unicode_literals

import hashlib
import sys


__all__ = ['PY3', 'string_types', 'int_types', 'sha1', 'unicode_type',
           'StringIO', 'BytesIO']


PY3 = sys.version_info >= (3, 0)


if not PY3:
    string_types = basestring
    int_types = int, long
    unicode_type = unicode

    from StringIO import StringIO
    BytesIO = StringIO
else:
    string_types = str
    int_types = int
    unicode_type = str

    from io import StringIO, BytesIO


class sha1(object):
    def __init__(self, arg=b''):
        self._hash = hashlib.sha1()
        if arg:
            self.update(arg)

    def update(self, arg):
        if not isinstance(arg, bytes):
            arg = arg.encode('ascii')
        self._hash.update(arg)

    def hexdigest(self):
        return self._hash.hexdigest()



from __future__ import division, unicode_literals

import json
import locale
import os
import sys
import tempfile
import warnings

from file_archive import FileStore, copy_file, BufferedReader
from file_archive.compat import int_types, unicode_type
from file_archive.errors import UsageWarning
from file_archive.trans import _


def parse_query_metadata(args):
    """Parses a list of key=value arguments or a hash value.

    Returns (hash:str, metadata:dict)
    """
    if len(args) == 1 and '=' not in args[0]:
        return args[0], None
    else:
        metadata = {}
        for a in args:
            k = a.split('=', 1)
            if len(k) != 2:
                sys.stderr.write(_("Metadata should have format key=value, "
                                   "key=type:value (eg. age=int:23) or "
                                   "key=type:req (eg. age=int:>21)\n"))
                sys.exit(1)
            k, v = k
            if ':' in v:
                t, v = v.split(':', 1)
                if t == 'int':
                    if v[0] == '>':
                        req, v = 'gt', v[1:]
                    elif v[0] == '<':
                        req, v = 'lt', v[1:]
                    else:
                        req = 'equal'
                    v = int(v)
                elif t == 'str':
                    req = 'equal'
                else:
                    sys.stderr.write(_("Metadata has unknown type '{t}'! "
                                       "Only 'str' and 'int' are supported.\n"
                                       "If you meant a string with a ':', "
                                       "use 'str:mystring'", t=t))
                    sys.exit(1)
            else:
                t = 'str'
                req = 'equal'
            if t == 'str' and isinstance(v, bytes):
                v = v.decode(locale.getpreferredencoding())
            if k in metadata:
                if t != metadata[k]['type']:
                    sys.stderr.write(_("Differing types for conditions on "
                                       "key {k}: {t1}, {t2}\n",
                                       k=k, t1=metadata[k]['type'], t2=t))
                    sys.exit(1)
                if req in metadata[k]:
                    sys.stderr.write(_("Multiple conditions {cond} on key "
                                       "{k}\n",
                                       cond=req, k=k))
                    sys.exit(1)
                metadata[k][req] = v
            else:
                metadata[k] = {'type': t, req: v}
        return None, metadata


def parse_new_metadata(args):
    """Parses a list of key=value or key=type:value arguments.
    """
    metadata = {}
    for a in args:
        k = a.split('=', 1)
        if len(k) != 2:
            sys.stderr.write(_("Metadata should have format key=value or "
                               "key=type:value (eg. age=int:23)\n"))
            sys.exit(1)
        k, v = k
        if ':' in v:
            t, v = v.split(':', 1)
        else:
            t = 'str'
        if k in metadata:
            sys.stderr.write("Multiple values for key %s\n" % k)
            sys.exit(1)
        if t == 'int':
            v = int(v)
        elif t == 'str':
            if isinstance(v, bytes):
                v = v.decode(locale.getpreferredencoding())
        else:
            sys.stderr.write(_("Metadata has unknown type '{t}'! Only 'str' "
                               "and 'int' are supported.\n"
                               "If you meant a string with a ':', use "
                               "'str:mystring'", t=t))
            sys.exit(1)
        metadata[k] = {'type': t, 'value': v}
    return metadata


def cmd_add(store, args):
    """Add command.

    add <filename> [key1=value1] [...]
    """
    if not args:
        sys.stderr.write(_("Missing filename\n"))
        sys.exit(1)
    filename = args[0]
    if not os.path.exists(filename):
        sys.stderr.write(_("Path does not exist: {path}\n", path=filename))
        sys.exit(1)
    metadata = parse_new_metadata(args[1:])
    if os.path.isdir(filename):
        entry = store.add_directory(filename, metadata)
    else:
        entry = store.add_file(filename, metadata)
    sys.stdout.write('%s\n' % entry.objectid)


def cmd_write(store, args):
    """Write command.

    write [key1=value1] [...]
    """
    metadata = parse_new_metadata(args)
    fd, filename = tempfile.mkstemp()
    os.close(fd)
    try:
        copy_file(sys.stdin, filename)
        entry = store.add_file(filename, metadata)
    finally:
        os.remove(filename)
    sys.stdout.write('%s\n' % entry['hash'])


def cmd_query(store, args):
    """Query command.

    query [-d] [-t] [key1=value1] [...]
    """
    pydict = False
    types = False
    while args and args[0][0] == '-':
        if args[0] == '-d':
            pydict = True
        elif args[0] == '-t':
            types = True
        elif args[0] == '--':
            del args[0]
            break
        else:
            sys.stderr.write(_("Unknown option: {opt}\n", opt=args[0]))
            sys.exit(1)
        del args[0]
    h, metadata = parse_query_metadata(args)
    if h is not None:
        entries = [store.get(h)]
    else:
        entries = store.query(metadata)

    if not pydict:
        for entry in sorted(entries, key=lambda e: e.objectid):
            sys.stdout.write("%s\n" % entry.objectid)
            for k, v in sorted(entry.metadata.items(), key=lambda p: p[0]):
                if types:
                    if isinstance(v, int_types):
                        v = 'int:%d' % v
                    else:  # isinstance(v, string_types):
                        v = 'str:%s' % v
                sys.stdout.write("\t%s\t%s\n" % (k, v))
    else:
        sys.stdout.write('{')
        for entry_nb, entry in enumerate(sorted(entries,
                                                key=lambda e: e.objectid)):
            sys.stdout.write(',\n' if entry_nb > 0 else '\n')
            sys.stdout.write('    "%s": {' % entry.objectid)
            for meta_nb, (k, v) in enumerate(sorted(entry.metadata.items(),
                                                    key=lambda p: p[0])):
                sys.stdout.write(',\n' if meta_nb > 0 else '\n')
                if types:
                    if isinstance(v, int_types):
                        v = '{"type": "int", "value": %d}' % v
                    else:  # isinstance(v, string_types):
                        assert isinstance(v, unicode_type)
                        v = '{"type": "str", "value": %s}' % json.dumps(v)
                else:
                    if isinstance(v, int_types):
                        v = '%d' % v
                    else:  # isinstance(v, string_types):
                        assert isinstance(v, unicode_type)
                        v = json.dumps(v)
                k = json.dumps(k)
                sys.stdout.write("        %s: %s" % (k, v))
            sys.stdout.write('\n    }')
        sys.stdout.write('\n}\n')


def cmd_print(store, args):
    """Print command.

    print [-m] [-t] <filehash> [...]
    print [-m] [-t] [key1=value1] [...]
    """
    meta = False
    types = False
    while args and args[0][0] == '-':
        if args[0] == '-m':
            meta = True
        elif args[0] == '-t':
            types = True
        elif args[0] == '--':
            del args[0]
            break
        else:
            sys.stderr.write(_("Unknown option: {opt}\n", opt=args[0]))
            sys.exit(1)
        del args[0]
    h, metadata = parse_query_metadata(args)
    if h is not None:
        try:
            entry = store.get(h)
        except KeyError:
            sys.stderr.write(_("Objectid not found\n"))
            sys.exit(2)
    else:
        entries = store.query(metadata)
        try:
            entry = next(entries)
        except StopIteration:
            sys.stderr.write(_("No match found\n"))
            sys.exit(2)
        try:
            next(entries)
        except StopIteration:
            pass
        else:
            sys.stderr.write(_("Warning: more matching files exist\n"))
    if meta:
        for k, v in entry.metadata.items():
            if k == 'hash':
                continue
            if types:
                if isinstance(v, int_types):
                    v = 'int:%d' % v
                else:  # isinstance(v, string_types):
                    v = 'str:%s' % v
            sys.stdout.write("%s\t%s\n" % (k, v))
    else:
        if os.path.isdir(entry.filename):
            sys.stderr.write(_("Error: match found but is a directory\n"))
            sys.exit(2)
        fp = entry.open()
        try:
            for chunk in BufferedReader(fp):
                sys.stdout.write(chunk)
        finally:
            fp.close()


def cmd_remove(store, args):
    """Remove command.

    remove [-f] <filehash>
    remove [-f] <key1=value1> [...]
    """
    if args and args[0] == '-f':
        del args[0]
        force = True
    else:
        force = False
    h, metadata = parse_query_metadata(args)
    if h is not None:
        store.remove(h)
    else:
        entries = store.query(metadata)
        if not args and not force:
            nb = sum(1 for e in entries)
            if nb:
                sys.stderr.write(_(
                        "Error: not removing files unconditionally unless -f "
                        "is given\n"
                        "(command would have removed {nb} files)\n",
                        nb=nb))
                sys.exit(1)
        for e in entries:
            store.remove(e)


def cmd_verify(store, args):
    """Verify command.

    This command accepts no argument.
    """
    if args:
        sys.stderr.write(_("verify command accepts no argument\n"))
        sys.exit(1)
    store.verify()


def cmd_view(store, args):
    if args:
        sys.stderr.write(_("view command accepts no argument\n"))
        sys.exit(1)
    from file_archive.viewer import run_viewer
    run_viewer(store)


commands = {
        'add': cmd_add,
        'write': cmd_write,
        'query': cmd_query,
        'print': cmd_print,
        'remove': cmd_remove,
        'verify': cmd_verify,
        'view': cmd_view}


def main(args):
    warnings.filterwarnings('always', category=UsageWarning)

    usage = _(
            "usage: {bin} <store> create\n"
            "   or: {bin} <store> add <filename> [key1=value1] [...]\n"
            "   or: {bin} <store> write [key1=value1] [...]\n"
            "   or: {bin} <store> query [-d] [-t] [key1=value1] [...]\n"
            "   or: {bin} <store> print [-m] [-t] <filehash> [...]\n"
            "   or: {bin} <store> print [-m] [-t] [key1=value1] [...]\n"
            "   or: {bin} <store> remove [-f] <filehash>\n"
            "   or: {bin} <store> remove [-f] <key1=value1> [...]\n"
            "   or: {bin} <store> verify\n"
            "   or: {bin} <store> view\n",
            bin='file_archive')

    if len(args) < 2:
        sys.stderr.write(usage)
        sys.exit(1)

    store = args[0]
    command = args[1]

    if command == 'create':
        try:
            FileStore.create_store(store)
        except Exception as e:
            sys.stderr.write(_("Can't create store: {err}\n", err=e.args[0]))
            sys.exit(3)
        sys.exit(0)

    try:
        store = FileStore(store)
    except Exception as e:
        sys.stderr.write(_("Invalid store: {err}\n", err=e.args[0]))
        sys.exit(3)

    try:
        try:
            func = commands[command]
        except KeyError:
            sys.stderr.write(usage)
            sys.exit(1)
        try:
            func(store, args[2:])
        except Exception:
            import traceback
            traceback.print_exc()
            sys.exit(3)
    finally:
        store.close()

    sys.exit(0)



from __future__ import division, unicode_literals

import platform
import subprocess
import sys

try:
    import sip

    api2_classes = [
            'QData', 'QDateTime', 'QString', 'QTextStream',
            'QTime', 'QUrl', 'QVariant']
    for cl in api2_classes:
        try:
            sip.setapi(cl, 2)
        except ValueError:
            pass

    from PyQt4 import QtCore, QtGui
except ImportError:
    sys.stderr.write("PyQt4 is required by 'file_archive view'\n")
    sys.exit(3)


try:
    import tdparser
except ImportError:
    sys.stderr.write("tdparser is required by 'file_archive view'\n")
    sys.exit(3)


from file_archive import hash_metadata
from file_archive.compat import int_types
from file_archive.parser import parse_expression
from file_archive.trans import _, _n


system = platform.system().lower()
if system == 'windows':
    def openfile(filename):
        subprocess.call(['cmd', '/c', 'start', filename])
elif system == 'darwin':
    def openfile(filename):
        subprocess.call(['open', filename])
elif system.startswith('linux'):
    def openfile(filename):
        subprocess.call(['xdg-open', filename])
else:
    openfile = None


__all__ = ['StoreViewerWindow', 'run_viewer']


class SearchError(Exception):
    """Error while querying the file store.
    """


class FileItem(QtGui.QTreeWidgetItem):
    def __init__(self, entry):
        self.entry = entry
        QtGui.QTreeWidgetItem.__init__(self, [entry.objectid])


class MetadataItem(FileItem):
    def __init__(self, entry, key, value):
        if isinstance(value, int_types):
            t = 'int'
            value = '%d' % value
        else:  # isinstance(v, string_types):
            t = 'str'
        QtGui.QTreeWidgetItem.__init__(self, [key, value, t])
        self.entry = entry


class StoreViewerWindow(QtGui.QMainWindow):
    WINDOW_TITLE = _("file_archive viewer")

    MAX_RESULTS = 100

    def __init__(self, store):
        QtGui.QMainWindow.__init__(self)
        self.setWindowTitle(self.WINDOW_TITLE)

        self.store = store

        searchbar = QtGui.QHBoxLayout()

        self._needs_refresh = False

        # Input line for the query
        self._input = QtGui.QLineEdit()
        self._input.setPlaceholderText(_("Enter query here"))
        self._input.returnPressed.connect(self._search)
        self._input.textEdited.connect(lambda t: self._set_needs_refresh())
        searchbar.addWidget(self._input)

        # Search button
        self._searchbutton = QtGui.QPushButton(_("Search"))
        self._searchbutton.clicked.connect(self._search)
        searchbar.addWidget(self._searchbutton)

        results = QtGui.QHBoxLayout()

        # Result view, as a tree with metadata
        self._result_tree = QtGui.QTreeWidget()
        self._result_tree.setColumnCount(3)
        self._result_tree.setHeaderLabels([_("Key"), _("Value"), _("Type")])
        self._result_tree.itemSelectionChanged.connect(self._selection_changed)
        results.addWidget(self._result_tree)

        # Buttons, enabled/disabled when the selection changes
        buttons = QtGui.QVBoxLayout()
        self._buttons = self._create_buttons()

        for name, button in self._buttons:
            buttons.addWidget(button)
        self._selection_changed()
        results.addLayout(buttons)

        layout = QtGui.QVBoxLayout()
        layout.addLayout(searchbar)
        layout.addLayout(results)

        widget = QtGui.QWidget()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        self._search()

    def _create_buttons(self):
        buttons = []

        # Open button; uses the system to choose the program to open with
        # (on Windows, might ask you what to use every time because of filename
        # scheme)
        open_button = QtGui.QPushButton(_("Open"))
        if openfile is not None:
            open_button.clicked.connect(self._openfile)
            buttons.append(('single', open_button))
        else:
            open_button.setEnabled(False)

        # Copy hash button
        copy_button = QtGui.QPushButton(_("Copy ID"))
        copy_button.clicked.connect(self._copy_objectid)
        buttons.append(('single', copy_button))

        # Edit metadata button
        edit_button = QtGui.QPushButton(_("Edit metadata..."))
        edit_button.clicked.connect(self._edit_metadata)
        buttons.append(('single', edit_button))

        # Delete button, removes what's selected (with confirmation)
        remove_button = QtGui.QPushButton(_("Delete"))
        remove_button.clicked.connect(self._delete)
        buttons.append(('multi', remove_button))

        return buttons

    def _set_needs_refresh(self, needs=True):
        if needs == self._needs_refresh:
            return

        if needs:
            self._searchbutton.setStyleSheet('font-weight: bold;')
        else:
            self._searchbutton.setStyleSheet('')
        self._needs_refresh = needs

    def _alter_search_conditions(self, conditions):
        return conditions

    def _search(self):
        error = None

        query = self._input.text()

        if len(query.split()) == 1 and all(o not in query.strip()
                                           for o in '=<>'):
            objectid = query.strip()
            try:
                entries = [self.store.get(objectid)]
            except KeyError:
                error = _("objectid '{oid}' not found", oid=objectid)

        else:
            try:
                conditions = parse_expression(query)
            except tdparser.Error as e:
                error = e.args[0]
            else:
                conditions = self._alter_search_conditions(conditions)
                entries = self.store.query(conditions, limit=self.MAX_RESULTS)

        self._result_tree.clear()

        if error is not None:
            w = QtGui.QTreeWidgetItem([error])
            w.setForeground(0, QtGui.QColor(255, 0, 0))
            self._result_tree.addTopLevelItem(w)
            self._result_tree.setFirstItemColumnSpanned(w, True)
        else:
            for i, entry in enumerate(entries):
                file_item = FileItem(entry)
                f = file_item.font(0)
                f.setBold(True)
                file_item.setFont(0, f)
                self._result_tree.addTopLevelItem(file_item)
                self._result_tree.setFirstItemColumnSpanned(file_item, True)
                for k, v in entry.metadata.items():
                    file_item.addChild(MetadataItem(entry, k, v))

                if i + 1 == self.MAX_RESULTS:
                    last_item = QtGui.QTreeWidgetItem(
                            [_("... stripped after {nb} results...",
                               nb=self.MAX_RESULTS)])
                    f = last_item.font(0)
                    f.setBold(True)
                    f.setItalic(True)
                    last_item.setFont(0, f)
                    self._result_tree.addTopLevelItem(last_item)
                    self._result_tree.setFirstItemColumnSpanned(last_item,
                                                                True)
                    break
            if self._result_tree.topLevelItemCount() == 0:
                w = QtGui.QTreeWidgetItem([_("No matches")])
                self._result_tree.addTopLevelItem(w)
                self._result_tree.setFirstItemColumnSpanned(w, True)
            self._result_tree.expandAll()

        self._set_needs_refresh(False)

    def _selection_changed(self):
        items = self._result_tree.selectedItems()
        for t, button in self._buttons:
            if t == 'single':
                button.setEnabled(
                        len(items) == 1 and
                        isinstance(items[0], FileItem))
            elif t == 'multi':
                button.setEnabled(bool(items) and
                                  all(isinstance(i, FileItem) for i in items))

    def _openfile(self):
        item = self._result_tree.currentItem()
        if item is not None:
            openfile(item.entry.filename)

    def _copy_objectid(self):
        items = self._result_tree.selectedItems()
        if not items:
            return
        objectid = items[0].entry.objectid

        clipboard = QtGui.QApplication.clipboard()
        clipboard.setText(objectid)

    def _edit_metadata(self):
        items = self._result_tree.selectedItems()
        if not items:
            return
        entry = items[0].entry

        editor = MetadataEditor(entry, self)
        editor.show()

    def change_metadata(self, old_objectid, metadata, remove_original=False):
        new_objectid = hash_metadata(metadata)
        if new_objectid == old_objectid:
            return

        self.store.metadata.add(new_objectid, metadata)

        if remove_original:
            self.store.remove(old_objectid)

        self._search()

    def _delete(self):
        items = self._result_tree.selectedItems()
        if not items:
            return
        confirm = QtGui.QMessageBox.question(
                self,
                _("Are you sure?"),
                _n("You are about to delete {num} entry from the store. "
                   "Please confirm.",
                   "You are about to delete {num} entries from the store. "
                   "Please confirm.",
                   len(items),
                   num=len(items)),
                QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel,
                QtGui.QMessageBox.Cancel)
        if confirm == QtGui.QMessageBox.Ok:
            ids = set([i.entry.objectid for i in items])
            i = 0
            while i < self._result_tree.topLevelItemCount():
                oid = self._result_tree.topLevelItem(i).entry.objectid
                if oid in ids:
                    self.store.remove(oid)
                    self._result_tree.takeTopLevelItem(i)
                else:
                    i += 1


class MetadataEditor(QtGui.QDialog):
    def __init__(self, entry, parent):
        QtGui.QDialog.__init__(self, parent, QtCore.Qt.Dialog)
        self.setWindowModality(QtCore.Qt.ApplicationModal)

        self._entry = entry
        self._parent = parent

        label = QtGui.QLabel(_("Editing entry %s") % entry.objectid)

        self._table = QtGui.QTableWidget()
        self._table.setColumnCount(3)
        self._table.setHorizontalHeaderLabels(['key', 'type', 'value'])
        self._table.setSortingEnabled(True)
        self._table.sortByColumn(0, QtCore.Qt.AscendingOrder)
        self._table.resizeColumnsToContents()
        scrollarea = QtGui.QScrollArea()
        scrollarea.setWidgetResizable(True)
        scrollarea.setWidget(self._table)

        plus = QtGui.QPushButton(_("+"))
        plus.clicked.connect(self._add_row)
        minus = QtGui.QPushButton(_("-"))
        minus.clicked.connect(self._remove_row)
        controls = QtGui.QVBoxLayout()
        controls.addWidget(plus)
        controls.addWidget(minus)

        table_row = QtGui.QHBoxLayout()
        table_row.addWidget(scrollarea)
        table_row.addLayout(controls)

        self._remove_original = QtGui.QCheckBox(_("Remove original entry"))
        self._remove_original.setChecked(False)
        self._remove_original.stateChanged.connect(self._mode_changed)

        self._ok_button = QtGui.QPushButton(_("Create new entry"))
        self._ok_button.clicked.connect(self._ok_clicked)
        cancel_button = QtGui.QPushButton(_("Cancel"))
        cancel_button.clicked.connect(lambda: self.setVisible(False))
        buttons = QtGui.QHBoxLayout()
        buttons.addStretch()
        buttons.addWidget(self._ok_button)
        buttons.addWidget(cancel_button)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(label)
        layout.addLayout(table_row)
        layout.addWidget(self._remove_original)
        layout.addLayout(buttons)
        self.setLayout(layout)

        self._table.setSortingEnabled(False)
        for k, v in self._entry.metadata.items():
            if k == 'hash':
                continue
            self._add_row(k, v, sorting_disabled=True)
        self._table.setSortingEnabled(True)

    def _add_row(self, key=None, value=None, sorting_disabled=False):
        row = self._table.rowCount()
        if not sorting_disabled:
            self._table.setSortingEnabled(False)
        self._table.insertRow(row)

        typeedit = QtGui.QComboBox()
        typeedit.addItems(['int', 'str'])
        self._table.setCellWidget(row, 1, typeedit)

        if key is not None:
            keyitem = QtGui.QTableWidgetItem(key)
            self._table.setItem(row, 0, keyitem)
        if value is not None:
            if isinstance(value, int_types):
                typeedit.setCurrentIndex(0)
                value = '%s' % value
            else:
                typeedit.setCurrentIndex(1)
            valueitem = QtGui.QTableWidgetItem(value)
            self._table.setItem(row, 2, valueitem)

        if not sorting_disabled:
            self._table.setSortingEnabled(True)

    def _remove_row(self):
        row = self._table.currentRow()
        self._table.removeRow(row)

    def _mode_changed(self, remove_original):
        if remove_original:
            self._ok_button.setText(_("Replace entry"))
        else:
            self._ok_button.setText(_("Create new entry"))

    def _ok_clicked(self):
        error = None

        metadata = {}

        for row in range(self._table.rowCount()):
            key = self._table.item(row, 0).text()
            type_ = self._table.cellWidget(row, 1).currentText()
            value = self._table.item(row, 2).text()

            if not key:
                error = _("Empty key")

            if type_ == 'int':
                try:
                    value = int(value)
                except ValueError:
                    error = (_("Invalid int value for %(key)s (row %(row)d)") %
                             {'key': key, 'row': row + 1})
            elif type_ != 'str':
                error = _("Invalid type (row %d)") % (row + 1)

            if key in metadata:
                error = _("Duplicate key %s" % key)

            if error is not None:
                break

            metadata[key] = {'type': type_, 'value': value}

        if error is not None:
            QtGui.QMessageBox.critical(self, _("Invalid values"), error)
            return

        metadata['hash'] = self._entry.metadata['hash']

        self._parent.change_metadata(self._entry.objectid, metadata,
                                     self._remove_original.isChecked())
        self.setVisible(False)


def run_viewer(store):
    application = QtGui.QApplication([])

    window = StoreViewerWindow(store)
    window.show()

    application.exec_()
    sys.exit(0)



from __future__ import division, unicode_literals

from tdparser import Lexer, Parser, Token, ParserError
from tdparser.topdown import EndToken


__all__ = ['parse_expression', 'parse_expressions']


def escape_string(text):
    return '"%s"' % text.replace('\\', '\\\\').replace('"', '\\"')


def unescape_string(text):
    assert text[0] == text[-1] == '"'
    return text[1:-1].replace('\\"', '"').replace('\\\\', '\\')


lexer = Lexer()


class Key(Token):
    regexp = r'[A-Za-z_][A-Za-z0-9_]*'

    def nud(self, context):
        return self


class Number(Token):
    regexp = r'-?\d+'
    type = 'int'

    def __init__(self, text):
        Token.__init__(self, text)
        self.value = int(text)

    def nud(self, context):
        return self


class String(Token):
    regexp = r'"(?:[^\\"]|\\\\|\\")*"'
    type = 'str'

    def __init__(self, text):
        Token.__init__(self, text)
        self.value = unescape_string(text)

    def nud(self, context):
        return self


class Operator(Token):
    regexp = r'[=<>]'
    lbp = 10

    def led(self, left, context):
        right = context.expression(self.lbp)
        inverted = False
        if not isinstance(left, Key) and isinstance(right, Key):
            left, right = right, left
            inverted = True
        elif not isinstance(left, Key):
            raise ParserError("Condition does not involve a key")
        elif isinstance(right, Key):
            raise ParserError("Condition involves two keys")
        if self.text == '=':
            cond = 'equal'
        elif self.text == '<':
            cond = 'gt' if inverted else 'lt'
        else:  # self.text == '>':
            cond = 'lt' if inverted else 'gt'
        return left.text, {'type': right.type, cond: right.value}


class ExistType(Token):
    regexp = r':'
    lbp = 10

    def led(self, left, context):
        right = context.expression(self.lbp)
        if not isinstance(left, Key):
            raise ParserError("Typed existence condition does not involve a "
                              "key")
        elif not isinstance(right, Key) or right.text not in ('int', 'str'):
            raise ParserError("Existence condition missing type")
        return left.text, {'type': right.text}


lexer.register_tokens(Key, Number, String, Operator, ExistType)


def _update_conditions(conditions, expr):
    key, cond = expr
    prec = conditions.get(key)
    if prec is not None:
        if prec['type'] != cond['type']:
            raise ParserError("Differing types for conditions on key %s: "
                              "%s, %s" % (key, prec['type'], cond['type']))
        for k in cond.keys():
            if k != 'type' and k in prec:
                raise ParserError("Multiple conditions %s on key %s" % (
                                  k, key))
        prec.update(cond)
    else:
        conditions[key] = cond


def parse_expression(expression):
    tokens = lexer.lex(expression)
    parser = Parser(tokens)
    conditions = {}
    while not isinstance(parser.current_token, EndToken):
        expr = parser.expression()
        if isinstance(expr, Token):
            raise ParserError("Found unexpected token %s in query" % expr)
        _update_conditions(conditions, expr)
    return conditions


def parse_expressions(expression_list):
    conditions = {}
    for expression in expression_list:
        expr = lexer.parse(expression)
        if isinstance(expr, Token):
            raise ParserError("Found unexpected token %s in query" % expr)
        _update_conditions(conditions, expr)
    return conditions



from __future__ import division, unicode_literals

import os
import shutil
import warnings

from file_archive.compat import string_types, sha1
from file_archive.database import normalize_metadata, MetadataStore
from file_archive.errors import CreationError, InvalidStore, UsageWarning


__version__ = '0.6'

__all__ = ['FileStore']


CHUNKSIZE = 4096


def BufferedReader(fp):
    """Generator that gives out chunks of the file.
    """
    chunk = fp.read(CHUNKSIZE)
    if not chunk:
        return
    yield chunk
    while len(chunk) == CHUNKSIZE:
        chunk = fp.read(CHUNKSIZE)
        if not chunk:
            return
        yield chunk


def hash_file(f):
    """Hashes a file to a 40 hex character SHA1.
    """
    h = sha1()
    h.update(b'file\n')
    for chunk in BufferedReader(f):
        h.update(chunk)
    return h.hexdigest()


def copy_file(fileobj, destination):
    """Copies a file object to a destination name.
    """
    with open(destination, 'wb') as destobj:
        try:
            for chunk in BufferedReader(fileobj):
                destobj.write(chunk)
        except:  # pragma: no cover
            os.remove(destination)
            raise


def relativize_link(link, root):
    """Tries to make the file a relative link.

    If the target is not inside root, returns None.
    root must be a realpath (os.path.realpath()).
    """
    target = os.path.join(os.path.dirname(link), os.readlink(link))
    target = os.path.realpath(target)
    root = os.path.realpath(root) + os.path.sep
    if os.path.commonprefix([target, root]) == root:
        return os.path.relpath(target, os.path.realpath(os.path.dirname(link)))
    else:
        return None


def hash_directory(path, root=None, visited=None):
    """Hashes a directory to a 40 hex character string.
    """
    h = sha1()
    if visited is None:
        visited = set()
    if os.path.realpath(path) in visited:
        raise ValueError("Can't hash directory structure: loop detected at "
                         "%s" % path)
    visited.add(os.path.realpath(path))
    if root is None:
        root = os.path.realpath(path)
    h.update(b'dir\n')
    for f in sorted(os.listdir(path)):
        pf = os.path.join(path, f)
        if os.path.islink(pf):
            link = relativize_link(pf, root)
            if link is not None:
                h.update('link %s %s\n' % (f, sha1(link).hexdigest()))
                continue
        if os.path.isdir(pf):
            if os.path.islink(pf):
                warnings.warn("%s is a symbolic link, recursing on target "
                              "directory" % pf,
                              UsageWarning)
            h.update('dir %s %s\n' % (f, hash_directory(pf, root, visited)))
        else:
            if os.path.islink(pf):
                warnings.warn("%s is a symbolic link, using target file "
                              "instead" % pf,
                              UsageWarning)
            with open(pf, 'rb') as fd:
                h.update('file %s %s\n' % (f, hash_file(fd)))
    return h.hexdigest()


def hash_metadata(metadata):
    """Hashes a dictionary of metadata.
    """
    assert 'hash' in metadata

    metadata = normalize_metadata(metadata)
    h = sha1()
    for k, v in sorted(metadata.items(), key=lambda p: p[0]):
        h.update('%d:%s' % (len(k), k))
        if v['type'] == 'int':
            h.update('i%de' % v['value'])
        else:  # v['type'] == 'str':
            h.update('%d:%s' % (len(v['value']), v['value']))

    return h.hexdigest()


def copy_directory(sourcepath, destination, root=None):
    """Copies a directory recursively to a destination name.
    """
    if root is None:
        root = os.path.realpath(sourcepath)
    # We don't display a warning for links, hash_directory() does that
    try:
        os.mkdir(destination)
        for f in os.listdir(sourcepath):
            pf = os.path.join(sourcepath, f)
            df = os.path.join(destination, f)
            if os.path.islink(pf):
                link = relativize_link(pf, root)
                if link is not None:
                    os.symlink(link, df)
                    continue
            if os.path.isdir(pf):
                copy_directory(pf, df, root)
            else:
                with open(pf, 'rb') as fd:
                    copy_file(fd, df)
    except:  # pragma: no cover
        shutil.rmtree(destination)
        raise


class Entry(object):
    """Represents a file in the store, along with its metadata.

    Use open() to get a file object, or use entry['somekey'] to read metadata
    values (also available as the entry.metadata dict).

    The metadata always contains at least 'hash', the hash of the file or
    directory associated with the entry.
    """
    def __init__(self, store, objectid, metadata):
        self.objectid = objectid
        self.filename = store._make_filename(metadata['hash'])
        self.metadata = metadata

    def __getitem__(self, key):
        return self.metadata[key]

    def open(self, binary=True):
        return open(self.filename, 'rb' if binary else 'r')


class EntryIterator(object):
    """Iterator returned by query().

    You can iterate on this to read all matching entries.
    """
    def __init__(self, store, infos):
        self.store = store
        self.metadata_iterator = infos

    def __iter__(self):
        return self

    def next(self):
        objectid, metadata = next(self.metadata_iterator)
        return Entry(self.store, objectid, metadata)
    __next__ = next


class FileStore(object):
    """Represents a file store.
    """
    def __init__(self, path):
        self.store = os.path.join(path, 'objects')
        if not os.path.isdir(self.store):
            raise InvalidStore("objects is not a directory")
        db = os.path.join(path, 'database')
        if not os.path.isfile(db):
            raise InvalidStore("database is not a file")
        self.metadata = MetadataStore(db)

    @staticmethod
    def create_store(path):
        if os.path.exists(path):
            if not os.path.isdir(path) or os.listdir(path):
                raise CreationError("Path is not a directory or is not empty")
            exists = True
        else:
            exists = False
        try:
            if not exists:
                os.mkdir(path)
            os.mkdir(os.path.join(path, 'objects'))
        except OSError as e:  # pragma: no cover
            raise CreationError("Could not create directories: %s: %s" % (
                    e.__class__.__name__, e.message))
        MetadataStore.create_db(os.path.join(path, 'database'))

    def close(self):
        self.metadata.close()
        self.metadata = None
        self.store = None

    def open_file(self, objectid, path=None, binary=True):
        """Returns a file object for a given objectid.
        """
        filepath = self.get_filename(objectid)
        if os.path.isdir(filepath):
            if path:
                return open(os.path.join(filepath, path),
                            'rb' if binary else 'r')
            else:
                raise ValueError("Object is a directory, not a file")
        else:  # os.path.isfile(filepath):
            if path is not None:
                raise ValueError("Object is a file, not a directory")
            else:
                return open(filepath, 'rb' if binary else 'r')

    def get_filename(self, objectid):
        """Returns the file path for a given objectid.
        """
        if not isinstance(objectid, string_types):
            raise TypeError("hash should be a string, not %s" % type(objectid))
        metadata = self.metadata.get(objectid)
        return self._make_filename(metadata['hash'])

    def _make_filename(self, filehash, make_dir=False):
        """Gets or makes the path for the given filehash.
        """
        dirname = os.path.join(self.store, filehash[:2])
        if make_dir and not os.path.isdir(dirname):
            os.mkdir(dirname)
        return os.path.join(dirname, filehash[2:])

    def add_file(self, newfile, metadata):
        """Adds a file given a file object or path and dict of metadata.

        The file will be copied/written in the store, and an entry will be
        added to the database.

        Note that, if you pass a file object, it needs to support
        newfile.seek(0, os.SEEK_SET) as it will be read twice: once to compute
        its SHA1 hash, and a second time to write it to disk.
        """
        if isinstance(newfile, string_types):
            if os.path.islink(newfile):
                warnings.warn("%s is a symbolic link, using target file "
                              "instead" % newfile,
                              UsageWarning)
            with open(newfile, 'rb') as fp:
                return self.add_file(fp, metadata)
        newfile.seek(0, os.SEEK_SET)
        filehash = hash_file(newfile)
        newfile.seek(0, os.SEEK_SET)
        metadata = dict(metadata)
        metadata['hash'] = filehash
        objectid = hash_metadata(metadata)
        storedfile = self._make_filename(filehash, make_dir=True)
        if not os.path.exists(storedfile):
            copy_file(newfile, storedfile)
        try:
            self.metadata.add(objectid, metadata)
        except:  # pragma: no cover
            os.remove(storedfile)
            raise
        return Entry(self, objectid, metadata)

    def add_directory(self, newdir, metadata):
        """Adds a directory given a path and dict of metadata.

        The directory will be recursively copied to the store, and an entry
        will be added to the database.
        """
        if not isinstance(newdir, string_types):
            raise TypeError("newdir should be a string, not %s" % type(newdir))
        try:
            dirhash = hash_directory(newdir)
        except (IOError, OSError):
            raise ValueError("Can't access directory")
        metadata = dict(metadata)
        metadata['hash'] = dirhash
        objectid = hash_metadata(metadata)
        storeddir = self._make_filename(dirhash, make_dir=True)
        if not os.path.exists(storeddir):
            copy_directory(newdir, storeddir)
        try:
            self.metadata.add(objectid, metadata)
        except:  # pragma: no cover
            shutil.rmtree(storeddir)
            raise
        return Entry(self, objectid, metadata)

    def add(self, newpath, metadata):
        """Adds a file or directory with a dict of metadata.

        This simply calls either add_file() or add_directory() with the given
        arguments.
        """
        if not isinstance(newpath, string_types):
            raise TypeError("newpath should be a string, not %s" %
                            type(newpath))
        if os.path.isdir(newpath):
            return self.add_directory(newpath, metadata)
        else:
            return self.add_file(newpath, metadata)

    def remove(self, objectid):
        """Removes a file or directory given its objectid.

        It is deleted from the store and removed from the database.
        """
        if isinstance(objectid, Entry):
            entry = objectid
        else:
            entry = self.get(objectid)
        self.metadata.remove(entry.objectid)
        if not self.metadata.has_filehash(entry['hash']):
            # Garbage collection
            if os.path.isdir(entry.filename):
                shutil.rmtree(entry.filename)
            else:
                os.remove(entry.filename)

    def get(self, objectid):
        """Gets an Entry from a hash.
        """
        metadata = self.metadata.get(objectid)  # Might raise KeyError
        return Entry(self, objectid, metadata)

    def query_one(self, conditions):
        """Returns at most one Entry matching the conditions.

        Returns one of the Entry object matching the conditions or None.
        """
        objectid, metadata = self.metadata.query_one(conditions)
        if objectid is None:
            return None
        else:
            return Entry(self, objectid, metadata)

    def query(self, conditions, limit=None):
        """Returns all the Entries matching the conditions.

        An EntryIterator is returned, with which you can access the different
        results.
        """
        infos = self.metadata.query_all(conditions, limit)
        return EntryIterator(self, infos)

    def verify(self):
        """Checks the integrity of the store.
        """
        # TODO : verify



from __future__ import division, unicode_literals

import sqlite3

from file_archive.compat import PY3, string_types, int_types
from file_archive.errors import Error, CreationError, InvalidStore


__all__ = ['MetadataStore']


if not PY3:
    class Row(sqlite3.Row):
        """Version of sqlite3.Row that doesn't choke on unicode column names.
        """
        def __getitem__(self, idx):  # pragma: no cover
            if isinstance(idx, unicode):
                idx = idx.encode('ascii')
            return sqlite3.Row.__getitem__(self, idx)
else:
    Row = sqlite3.Row


def normalize_metadata(metadata):
    result = {}
    for mkey, mvalue in metadata.items():
        if isinstance(mvalue, string_types):
            t = 'str'
        elif isinstance(mvalue, int_types):
            t = 'int'
        elif isinstance(mvalue, dict):
            r = dict(mvalue)
            try:
                t = r.pop('type')
                mvalue = r.pop('value')
                if r:
                    raise KeyError
            except KeyError:
                raise ValueError("Metadata values should be "
                                 "dictionaries with the format:\n"
                                 "{'type': 'int/str/...', "
                                 "'value': <value>}")
        else:
            raise TypeError(
                    "Metadata values should be int, string, or dictionaries "
                    "with the format:\n"
                    "{'type': 'int/str/...', 'value': <value>}")
        if isinstance(mvalue, bytes):
            mvalue = mvalue.decode('ascii')
        result[mkey] = {'type': t, 'value': mvalue}
    return result


class MetadataStore(object):
    """The database holding metadata associated to SHA1 hashs.
    """
    _TYPES = [('TEXT', 'str'), ('INTEGER', 'int')]

    def __init__(self, database):
        try:
            self.conn = sqlite3.connect(database)
            self.conn.row_factory = Row
            cur = self.conn.cursor()
            tables = cur.execute('''
                    SELECT name FROM sqlite_master WHERE type = 'table'
                    ''')
            tables = set(r['name'] for r in tables.fetchall())
            if tables != set(['metadata']):
                raise InvalidStore("Database doesn't have required structure")
        except sqlite3.Error as e:
            raise InvalidStore("Cannot access database: %s: %s" % (
                    e.__class__.__name__, e.message))

    @staticmethod
    def create_db(database):
        try:
            conn = sqlite3.connect(database)
            query = '''
                    CREATE TABLE metadata(
                        objectid VARCHAR(40) NOT NULL,
                        mkey VARCHAR(255) NULL
                    '''
            indexes = [
                    'CREATE INDEX id_idx ON metadata(objectid)',
                    'CREATE INDEX mkey_idx ON metadata(mkey)']

            for datatype, name in MetadataStore._TYPES:
                query += '''
                        , mvalue_{name} {type} NULL
                        '''.format(name=name, type=datatype)
                indexes.append('''
                        CREATE INDEX mvalue_{name} ON metadata(mvalue_{name})
                        '''.format(name=name))
            query += ')'

            cur = conn.cursor()
            cur.execute(query)
            for idx_query in indexes:
                cur.execute(idx_query)

            conn.commit()
            conn.close()
        except sqlite3.Error as e:  # pragma: no cover
            raise CreationError("Could not create database: %s: %s" % (
                    e.__class__.__name__, e.message))

    def close(self):
        self.conn.commit()
        self.conn.close()

    def add(self, objectid, metadata):
        """Adds an object to the store.

        Returns True if it wasn't already stored.
        """
        assert 'hash' in metadata
        metadata = normalize_metadata(metadata)
        cur = self.conn.cursor()
        try:
            cur.execute(
                    '''
                    SELECT objectid FROM metadata
                    WHERE objectid = :objectid
                    LIMIT 1
                    ''',
                    {'objectid': objectid})
            if cur.fetchone() is not None:
                return False
            for mkey, mvalue in metadata.items():
                t = mvalue['type']
                v = mvalue['value']
                cur.execute(
                        '''
                        INSERT INTO metadata(objectid, mkey, mvalue_{name})
                        VALUES(:objectid, :key, :value)
                        '''.format(name=t),
                        {'objectid': objectid, 'key': mkey, 'value': v})
            self.conn.commit()
            return True
        except:
            self.conn.rollback()
            raise

    def remove(self, objectid):
        """Removes an object from the store.

        Raises KeyError if the entry didn't exist.
        """
        cur = self.conn.cursor()
        try:
            cur.execute(
                    '''
                    DELETE FROM metadata WHERE objectid = :objectid
                    ''',
                    {'objectid': objectid})
            if not cur.rowcount:
                raise KeyError(objectid)
            self.conn.commit()
        except:
            self.conn.rollback()
            raise

    def get(self, objectid):
        """Gets an entry from its objectid, as a dict.
        """
        cur = self.conn.cursor()
        rows = cur.execute(
                '''
                SELECT * FROM metadata
                WHERE objectid = :objectid
                ''',
                {'objectid': objectid})
        result = ResultBuilder(rows)
        try:
            objectid_, metadata = next(result)
            return metadata
        except StopIteration:
            raise KeyError("No entry with this objectid")

    def has_filehash(self, filehash):
        """Checks for at least one entry with the given file hash.

        File should be garbage-collected if no entry refers to it.
        """
        cur = self.conn.cursor()
        rows = cur.execute(
                '''
                SELECT * FROM metadata
                WHERE mkey = 'hash' AND mvalue_str = :filehash
                ''',
                {'filehash': filehash})
        try:
            next(rows)
            return True
        except StopIteration:
            return False

    def query_one(self, conditions):
        """Returns at most one entry matching the conditions, as a dict.

        Returns objectid, metadata:dict

        The metadata dict will have the 'hash' key plus all the stored
        metadata.

        `conditions` is a dictionary of metadata that need to be included in
        the actual dict of each entry.
        """
        rows = self.query_all(conditions, limit=1)
        try:
            return next(rows)
        except StopIteration:
            return None, None

    def query_all(self, conditions, limit=None):
        """Returns an iterable of rows matching the conditions.

        Each row is a pair (objectid, metadata), where metadata will have the
        'hash' key plus all the stored metadata.

        `conditions` is a dictionary of metadata that need to be included in
        the actual dict of each entry.
        """
        # Build the LIMIT part from the limit arg (number or None)
        if limit is not None:
            limit = 'LIMIT %d' % limit
        else:
            limit = ''

        cur = self.conn.cursor()
        if not conditions:
            hquery = '''
                    SELECT DISTINCT objectid
                    FROM metadata
                    {limit}
                    '''.format(limit=limit)
            params = {}
        else:
            conditems = self._make_conditions(conditions)
            i, key, cond0, params = next(conditems)
            hquery = '''
                    SELECT i0.objectid
                    FROM metadata i0
                    '''
            params['key0'] = key
            for i, key, cond, prms in conditems:
                hquery += '''
                        INNER JOIN metadata i{i} ON i0.objectid = i{i}.objectid
                            AND i{i}.mkey = :key{i} {cond}
                        '''.format(i=i, cond='AND ' + cond if cond else '')
                params['key%d' % i] = key
                params.update(prms)
            hquery += '''
                    WHERE i0.mkey = :key0 {cond}
                    {limit}
                    '''.format(cond='AND ' + cond0 if cond0 else '',
                               limit=limit)

        # And we put that in the query
        rows = cur.execute(
                '''
                SELECT *
                FROM metadata
                WHERE objectid IN ({ids})
                ORDER BY objectid
                '''.format(ids=hquery),
                params)

        return ResultBuilder(rows)

    def _make_conditions(self, conditions):
        for i, (key, value) in enumerate(conditions.items()):
            t = None
            if isinstance(value, string_types):
                t = 'str'
                req = [('equal', value)]
            elif isinstance(value, int_types):
                t = 'int'
                req = [('equal', value)]
            elif isinstance(value, dict) and not value:
                # Empty dict: key exist with any type or value
                req = value
            elif isinstance(value, dict):
                req = dict(value)
                try:
                    t = req.pop('type')
                except KeyError:
                    raise TypeError("Query conditions should include key "
                                    "'type'")
                req = iter(req.items())
                if t not in ('str', 'int'):
                    raise TypeError("Unknown data type %r" % t)
            else:
                raise TypeError(
                        "Query conditions should be dictionaries with the "
                        "format:\n"
                        "{'type': 'int/str/...', <condition>}")

            conds = []
            params = {}
            if req:
                var = 'i{i}.mvalue_{t}'.format(i=i, t=t)
                for j, (k, v) in enumerate(req):
                    val = ':val{i}_{j}'.format(i=i, j=j)
                    params['val%d_%d' % (i, j)] = v
                    if k == 'equal':
                        conds.append('{var} = {val}'.format(var=var, val=val))
                    elif t == 'int' and k == 'lt':
                        conds.append('{var} < {val}'.format(var=var, val=val))
                    elif t == 'int' and k == 'gt':
                        conds.append('{var} > {val}'.format(var=var, val=val))
                    else:
                        raise ValueError("Unsupported operation %r" % k)
            elif t is not None:
                # Just check type
                conds = ['{var} IS NOT NULL'.format(
                         var='i{i}.mvalue_{t}'.format(i=i, t=t))]
            yield (i,
                   key,
                   ' AND '.join(conds),
                   params)


class ResultBuilder(object):
    """This regroups rows for key-values of a single entry into one dict.

    Example:
    +--------+----+------+
    |objectid|mkey|mvalue|    [
    +--------+----+------+     'aaaa', {'hash': 'xxxx', 'one': 11, 'two': 12},
    |  aaaa  |one |  11  | =>  'bbbb', {'hash': 'yyyy', 'one': 21, 'six': 26},
    |  aaaa  |two |  12  |    ]
    |  bbbb  |one |  21  |
    |  bbbb  |six |  26  |
    +--------+----+------+
    """
    def __init__(self, rows):
        self.rows = iter(rows)
        self.record = None

    def __iter__(self):  # pragma: no cover
        return self

    def next(self):
        if self.rows is None:
            raise StopIteration
        if self.record is None:
            r = next(self.rows)  # Might raise StopIteration
        else:
            r = self.record
        objectid = r['objectid']

        def get_value(r):
            for datatype, name in MetadataStore._TYPES:
                v = r['mvalue_%s' % name]
                if v is not None:
                    return v
            else:  # pragma: no cover
                raise Error("SQL query didn't return a value for "
                            "objectid=%s, key=%s" % (r['objectid'], r['mkey']))

        # We are outer joining, so an objectid with no metadata will be
        # returned as a single row with mkey and everything but objectid NULL
        if len(r) > 1 and r['mkey']:
            dct = {r['mkey']: get_value(r)}
        else:
            dct = {}

        for r in self.rows:
            if r['objectid'] != objectid:
                self.record = r
                assert 'hash' in dct
                return objectid, dct
            dct[r['mkey']] = get_value(r)
        else:
            self.rows = None
        assert 'hash' in dct
        return objectid, dct
    __next__ = next



__all__ = ['Error', 'InvalidStore', 'CreationError', 'UsageWarning']


class Error(Exception):
    """Base class for this package's exceptions.
    """


class InvalidStore(ValueError, Error):
    """Indicates that the store at the given path is corrupted or not a store.
    """


class CreationError(IOError, Error):
    """Failed to create a new store at the given path.
    """


class UsageWarning(UserWarning):
    """Something unsafe was requested and carried out.
    """



import os
from setuptools import setup


# pip workaround
os.chdir(os.path.abspath(os.path.dirname(__file__)))


description = """
file_archive is a file archiving system. You submit it files with a set of
metadata, as key-value pairs, and it allows you to later retrieve the files
that match conditions on these metadata.

It uses a flat file-store where files are stored under their 40 characters SHA1
hash, and a SQLite3 database for the metadata.

Its purpose is to be used as a persistent file store for the VisTrails workflow
and provenance management system: http://www.vistrails.org/
"""
setup(name='file_archive',
      version='0.6',
      packages=['file_archive'],
      entry_points={
        'console_scripts': [
          'file_archive = file_archive.entry_point:entry_point']},
      description="A file store with searchable metadata",
      author="Remi Rampin",
      author_email='vistrails-dev@vistrails.org',
      url='http://github.com/remram44/file_archive',
      long_description=description,
      license='Modified BSD License',
      package_data={
        'file_archive': ['l10n/*/LC_MESSAGES/*.mo'],
      },
      zip_safe=True,
      keywords=['file', 'archive', 'metadata', 'vida', 'nyu'],
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: System :: Archiving'],
      install_requires=['tdparser>=1.1.4'])



from __future__ import division, unicode_literals

import codecs
import locale
import sys

from file_archive.main import main
from file_archive.trans import setup_translation


def entry_point():
    # Locale
    locale.setlocale(locale.LC_ALL, str(''))

    # Encoding for output streams
    if str == bytes:
        writer = codecs.getwriter(locale.getpreferredencoding())
        sys.stdout = writer(sys.stdout)
        sys.stderr = writer(sys.stderr)

    # Load gettext translation catalog
    setup_translation()

    main(sys.argv[1:])



from __future__ import division, unicode_literals

import gettext
import locale
import pkg_resources


__all__ = ['setup_translation', '_', '_n']


# Entry points should be using:
#   locale.setlocale(locale.LC_ALL, '')
#   file_archive.trans.setup_translation()


if str == bytes:
    string_types = basestring
else:
    string_types = str


trans = gettext.NullTranslations()


def setup_translation(languages=None):
    global trans

    if languages is None:
        languages = []
    elif isinstance(languages, string_types):
        languages = [languages]

    d = pkg_resources.resource_filename('file_archive', 'l10n')

    lang = locale.getlocale()[0]
    if lang is not None:
        languages.append(lang)
    trans = gettext.translation('file_archive', d, languages, fallback=True)


if hasattr(trans, 'ugettext'):
    def _(*args, **kwargs):
        tr = trans.ugettext(*args)
        if kwargs:
            tr = tr.format(**kwargs)
        return tr

    def _n(singular, plural, n, **kwargs):
        tr = trans.ungettext(singular, plural, n)
        if kwargs:
            tr = tr.format(**kwargs)
        return tr
else:
    # ugettext was removed from Python 3
    def _(*args, **kwargs):
        tr = trans.gettext(*args)
        if kwargs:
            tr = tr.format(**kwargs)
        return tr

    def _n(singular, plural, n, **kwargs):
        tr = trans.ngettext(singular, plural, n)
        if kwargs:
            tr = tr.format(**kwargs)
        return tr
